from rope.base import utils
from rope.base.oi.type_hinting import interfaces
from rope.base.oi.type_hinting.providers import (
    composite,
    docstrings,
    inheritance,
    numpydocstrings,
    pep0484_type_comments,
)
from rope.base.oi.type_hinting.resolvers import composite as composite_resolvers
from rope.base.oi.type_hinting.resolvers import types


class TypeHintingFactory(interfaces.ITypeHintingFactory):
    @utils.saveit
    def make_param_provider(self):
        providers = [
            docstrings.ParamProvider(
                docstrings.DocstringParamParser(), self.make_resolver()
            ),
            docstrings.ParamProvider(
                numpydocstrings.NumPyDocstringParamParser(), self.make_resolver()
            ),
        ]
        return inheritance.ParamProvider(composite.ParamProvider(*providers))

    @utils.saveit
    def make_return_provider(self):
        providers = [
            docstrings.ReturnProvider(
                docstrings.DocstringReturnParser(), self.make_resolver()
            ),
        ]
        return inheritance.ReturnProvider(composite.ReturnProvider(*providers))

    @utils.saveit
    def make_assignment_provider(self):
        providers = [
            pep0484_type_comments.AssignmentProvider(self.make_resolver()),
            docstrings.AssignmentProvider(
                docstrings.DocstringParamParser(), self.make_resolver()
            ),
            docstrings.AssignmentProvider(
                numpydocstrings.NumPyDocstringParamParser(), self.make_resolver()
            ),
        ]
        return inheritance.AssignmentProvider(composite.AssignmentProvider(*providers))

    @utils.saveit
    def make_resolver(self):
        """
        :rtype: rope.base.oi.type_hinting.resolvers.interfaces.IResolver
        """
        resolvers = [
            types.Resolver(),
        ]
        return composite_resolvers.Resolver(*resolvers)


default_type_hinting_factory = TypeHintingFactory()


class TypeHintingFactoryAccessor:
    def __call__(self, project):
        """
        :type project: rope.base.project.Project
        :rtype: rope.base.oi.type_hinting.interfaces.ITypeHintingFactory
        """
        factory_location = project.get_prefs().get(
            "type_hinting_factory",
            "rope.base.oi.type_hinting.factory.default_type_hinting_factory",
        )
        return self._get_factory(factory_location)

    @utils.cached(10)
    def _get_factory(self, factory_location):
        """
        :type factory_location: str
        :rtype: rope.base.oi.type_hinting.interfaces.ITypeHintingFactory
        """
        return utils.resolve(factory_location)


get_type_hinting_factory = TypeHintingFactoryAccessor()



import unittest
from textwrap import dedent

from rope.base import ast
from rope.refactor import suites
from ropetest import testutils


class SuiteTest(unittest.TestCase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_trivial_case(self):
        root = source_suite_tree("")
        self.assertEqual(1, root.get_start())
        self.assertEqual(0, len(root.get_children()))

    def test_simple_ifs(self):
        root = source_suite_tree(dedent("""\
            if True:
                pass"""))
        self.assertEqual(1, len(root.get_children()))

    def test_simple_else(self):
        root = source_suite_tree(dedent("""\
            if True:
                pass
            else:
                pass
        """))
        self.assertEqual(2, len(root.get_children()))
        self.assertEqual(1, root.get_children()[1].get_start())

    def test_for(self):
        root = source_suite_tree(dedent("""\

            for i in range(10):
                pass
            else:
                pass
        """))
        self.assertEqual(2, len(root.get_children()))
        self.assertEqual(2, root.get_children()[1].get_start())

    def test_while(self):
        root = source_suite_tree(dedent("""\
            while True:
                pass
        """))
        self.assertEqual(1, len(root.get_children()))
        self.assertEqual(1, root.get_children()[0].get_start())

    def test_with(self):
        root = source_suite_tree(dedent("""\
            from __future__ import with_statement
            with file(x):    pass
        """))
        self.assertEqual(1, len(root.get_children()))
        self.assertEqual(2, root.get_children()[0].get_start())

    def test_try_finally(self):
        root = source_suite_tree(dedent("""\
            try:
                pass
            finally:
                pass
        """))
        self.assertEqual(2, len(root.get_children()))
        self.assertEqual(1, root.get_children()[0].get_start())

    def test_try_except(self):
        root = source_suite_tree(dedent("""\
            try:
                pass
            except:
                pass
            else:
                pass
        """))
        self.assertEqual(3, len(root.get_children()))
        self.assertEqual(1, root.get_children()[2].get_start())

    def test_try_except_finally(self):
        root = source_suite_tree(dedent("""\
            try:
                pass
            except:
                pass
            finally:
                pass
        """))
        self.assertEqual(3, len(root.get_children()))
        self.assertEqual(1, root.get_children()[2].get_start())

    def test_local_start_and_end(self):
        root = source_suite_tree(dedent("""\
            if True:
                pass
            else:
                pass
        """))
        self.assertEqual(1, root.local_start())
        self.assertEqual(4, root.local_end())
        if_suite = root.get_children()[0]
        self.assertEqual(2, if_suite.local_start())
        self.assertEqual(2, if_suite.local_end())
        else_suite = root.get_children()[1]
        self.assertEqual(4, else_suite.local_start())
        self.assertEqual(4, else_suite.local_end())

    def test_find_suite(self):
        root = source_suite_tree("\n")
        self.assertEqual(root, root.find_suite(1))

    def test_find_suite_for_ifs(self):
        root = source_suite_tree(dedent("""\
            if True:
                pass
        """))
        if_suite = root.get_children()[0]
        self.assertEqual(if_suite, root.find_suite(2))

    def test_find_suite_for_between_suites(self):
        root = source_suite_tree(dedent("""\
            if True:
                pass
            print(1)
            if True:
                pass
        """))
        if_suite1 = root.get_children()[0]
        if_suite2 = root.get_children()[1]
        self.assertEqual(if_suite1, root.find_suite(2))
        self.assertEqual(if_suite2, root.find_suite(5))
        self.assertEqual(root, root.find_suite(3))

    def test_simple_find_visible(self):
        root = source_suite_tree("a = 1\n")
        self.assertEqual(1, suites.find_visible_for_suite(root, [1]))

    def test_simple_find_visible_ifs(self):
        root = source_suite_tree(dedent("""\

            if True:
                a = 1
                b = 2
        """))
        self.assertEqual(root.find_suite(3), root.find_suite(4))
        self.assertEqual(3, suites.find_visible_for_suite(root, [3, 4]))

    def test_simple_find_visible_for_else(self):
        root = source_suite_tree(dedent("""\

            if True:
                pass
            else:    pass
        """))
        self.assertEqual(2, suites.find_visible_for_suite(root, [2, 4]))

    def test_simple_find_visible_for_different_suites(self):
        root = source_suite_tree(dedent("""\
            if True:
                pass
            a = 1
            if False:
                pass
        """))
        self.assertEqual(1, suites.find_visible_for_suite(root, [2, 3]))
        self.assertEqual(5, suites.find_visible_for_suite(root, [5]))
        self.assertEqual(1, suites.find_visible_for_suite(root, [2, 5]))

    def test_not_always_selecting_scope_start(self):
        root = source_suite_tree(dedent("""\
            if True:
                a = 1
                if True:
                    pass
                else:
                    pass
        """))
        self.assertEqual(3, suites.find_visible_for_suite(root, [4, 6]))
        self.assertEqual(3, suites.find_visible_for_suite(root, [3, 5]))
        self.assertEqual(3, suites.find_visible_for_suite(root, [4, 5]))

    def test_ignoring_functions(self):
        root = source_suite_tree(dedent("""\
            def f():
                pass
            a = 1
        """))
        self.assertEqual(3, suites.find_visible_for_suite(root, [2, 3]))

    def test_ignoring_classes(self):
        root = source_suite_tree(dedent("""\
            a = 1
            class C():
                pass
        """))
        self.assertEqual(1, suites.find_visible_for_suite(root, [1, 3]))

    @testutils.only_for_versions_higher("3.10")
    def test_match_case(self):
        root = source_suite_tree(dedent("""\
            a = 1
            match var:
                case Foo("xx"):
                    print(x)
                case Foo(x):
                    print(x)
        """))
        self.assertEqual(root.find_suite(4), root.find_suite(6))
        self.assertEqual(root.find_suite(3), root.find_suite(6))
        self.assertEqual(2, suites.find_visible_for_suite(root, [2, 4]))


def source_suite_tree(source):
    return suites.ast_suite_tree(ast.parse(source))



import unittest
from textwrap import dedent

from rope.refactor import restructure
from ropetest import testutils


class RestructureTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_trivial_case(self):
        refactoring = restructure.Restructure(self.project, "a = 1", "a = 0")
        self.mod.write("b = 1\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual("b = 1\n", self.mod.read())

    def test_replacing_simple_patterns(self):
        refactoring = restructure.Restructure(self.project, "a = 1", "a = int(1)")
        self.mod.write("a = 1\nb = 1\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = int(1)\nb = 1\n", self.mod.read())

    def test_replacing_patterns_with_normal_names(self):
        refactoring = restructure.Restructure(
            self.project, "${a} = 1", "${a} = int(1)", args={"a": "exact"}
        )
        self.mod.write("a = 1\nb = 1\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = int(1)\nb = 1\n", self.mod.read())

    def test_replacing_patterns_with_any_names(self):
        refactoring = restructure.Restructure(self.project, "${a} = 1", "${a} = int(1)")
        self.mod.write("a = 1\nb = 1\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = int(1)\nb = int(1)\n", self.mod.read())

    def test_replacing_patterns_with_any_names2(self):
        refactoring = restructure.Restructure(self.project, "${x} + ${x}", "${x} * 2")
        self.mod.write("a = 1 + 1\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = 1 * 2\n", self.mod.read())

    def test_replacing_patterns_with_checks(self):
        self.mod.write(dedent("""\
            def f(p=1):
                return p
            g = f
            g()
        """))
        refactoring = restructure.Restructure(
            self.project, "${f}()", "${f}(2)", args={"f": "object=mod.f"}
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p=1):
                    return p
                g = f
                g(2)
            """),
            self.mod.read(),
        )

    def test_replacing_assignments_with_sets(self):
        refactoring = restructure.Restructure(
            self.project, "${a} = ${b}", "${a}.set(${b})"
        )
        self.mod.write("a = 1\nb = 1\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual("a.set(1)\nb.set(1)\n", self.mod.read())

    def test_replacing_sets_with_assignments(self):
        refactoring = restructure.Restructure(
            self.project, "${a}.set(${b})", "${a} = ${b}"
        )
        self.mod.write("a.set(1)\nb.set(1)\n")
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                a = 1
                b = 1
            """),
            self.mod.read(),
        )

    def test_using_make_checks(self):
        self.mod.write(dedent("""\
            def f(p=1):
                return p
            g = f
            g()
        """))
        refactoring = restructure.Restructure(
            self.project, "${f}()", "${f}(2)", args={"f": "object=mod.f"}
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p=1):
                    return p
                g = f
                g(2)
            """),
            self.mod.read(),
        )

    def test_using_make_checking_builtin_types(self):
        self.mod.write("a = 1 + 1\n")
        refactoring = restructure.Restructure(
            self.project, "${i} + ${i}", "${i} * 2", args={"i": "type=__builtin__.int"}
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = 1 * 2\n", self.mod.read())

    def test_auto_indentation_when_no_indentation(self):
        self.mod.write("a = 2\n")
        refactoring = restructure.Restructure(
            self.project, "${a} = 2", "${a} = 1\n${a} += 1"
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                a = 1
                a += 1
            """),
            self.mod.read(),
        )

    def test_auto_indentation(self):
        self.mod.write(dedent("""\
            def f():
                a = 2
        """))
        refactoring = restructure.Restructure(
            self.project, "${a} = 2", "${a} = 1\n${a} += 1"
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                def f():
                    a = 1
                    a += 1
            """),
            self.mod.read(),
        )

    def test_auto_indentation_and_not_indenting_blanks(self):
        self.mod.write("def f():\n    a = 2\n")
        refactoring = restructure.Restructure(
            self.project, "${a} = 2", "${a} = 1\n\n${a} += 1"
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual("def f():\n    a = 1\n\n    a += 1\n", self.mod.read())

    def test_importing_names(self):
        self.mod.write("a = 2\n")
        refactoring = restructure.Restructure(
            self.project, "${a} = 2", "${a} = myconsts.two", imports=["import myconsts"]
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                import myconsts
                a = myconsts.two
            """),
            self.mod.read(),
        )

    def test_not_importing_names_when_there_are_no_changes(self):
        self.mod.write("a = True\n")
        refactoring = restructure.Restructure(
            self.project, "${a} = 2", "${a} = myconsts.two", imports=["import myconsts"]
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = True\n", self.mod.read())

    def test_handling_containing_matches(self):
        self.mod.write("a = 1 / 2 / 3\n")
        refactoring = restructure.Restructure(
            self.project, "${a} / ${b}", "${a} // ${b}"
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = 1 // 2 // 3\n", self.mod.read())

    def test_handling_overlapping_matches(self):
        self.mod.write("a = 1\na = 1\na = 1\n")
        refactoring = restructure.Restructure(self.project, "a = 1\na = 1\n", "b = 1")
        self.project.do(refactoring.get_changes())
        self.assertEqual("b = 1\na = 1\n", self.mod.read())

    def test_preventing_stack_overflow_when_matching(self):
        self.mod.write("1\n")
        refactoring = restructure.Restructure(self.project, "${a}", "${a}")
        self.project.do(refactoring.get_changes())
        self.assertEqual("1\n", self.mod.read())

    def test_performing_a_restructuring_to_all_modules(self):
        mod2 = testutils.create_module(self.project, "mod2")
        self.mod.write("a = 1\n")
        mod2.write("b = 1\n")
        refactoring = restructure.Restructure(self.project, "1", "2 / 1")
        self.project.do(refactoring.get_changes())
        self.assertEqual("a = 2 / 1\n", self.mod.read())
        self.assertEqual("b = 2 / 1\n", mod2.read())

    def test_performing_a_restructuring_to_selected_modules(self):
        mod2 = testutils.create_module(self.project, "mod2")
        self.mod.write("a = 1\n")
        mod2.write("b = 1\n")
        refactoring = restructure.Restructure(self.project, "1", "2 / 1")
        self.project.do(refactoring.get_changes(resources=[mod2]))
        self.assertEqual("a = 1\n", self.mod.read())
        self.assertEqual("b = 2 / 1\n", mod2.read())

    def test_unsure_argument_of_default_wildcard(self):
        self.mod.write(dedent("""\
            def f(p):
                return p * 2
            x = "" * 2
            i = 1 * 2
        """))
        refactoring = restructure.Restructure(
            self.project,
            "${s} * 2",
            "dup(${s})",
            args={"s": {"type": "__builtins__.str", "unsure": True}},
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
                def f(p):
                    return dup(p)
                x = dup("")
                i = 1 * 2
            """),
            self.mod.read(),
        )

    def test_statement_after_string_and_column(self):
        mod_text = dedent("""\
            def f(x):
              if a == "a": raise Exception("test")
        """)
        self.mod.write(mod_text)
        refactoring = restructure.Restructure(self.project, "${a}", "${a}")
        self.project.do(refactoring.get_changes())
        self.assertEqual(mod_text, self.mod.read())

    @testutils.only_for_versions_higher("3.3")
    def test_yield_from(self):
        mod_text = dedent("""\
            def f(lst):
                yield from lst
        """)
        self.mod.write(mod_text)
        refactoring = restructure.Restructure(
            self.project,
            "yield from ${a}",
            dedent("""\
                for it in ${a}:
                   yield it"""),
        )
        self.project.do(refactoring.get_changes())
        self.assertEqual(
            dedent("""\
            def f(lst):
                for it in lst:
                   yield it
            """),
            self.mod.read(),
        )



import unittest
from textwrap import dedent

import rope.base.evaluate
from rope.base import codeanalyze, exceptions, libutils, worder
from rope.base.codeanalyze import LogicalLineFinder, SourceLinesAdapter, get_block_start
from ropetest import testutils


class SourceLinesAdapterTest(unittest.TestCase):
    def test_source_lines_simple(self):
        to_lines = SourceLinesAdapter("line1\nline2\n")
        self.assertEqual("line1", to_lines.get_line(1))
        self.assertEqual("line2", to_lines.get_line(2))
        self.assertEqual("", to_lines.get_line(3))
        self.assertEqual(3, to_lines.length())

    def test_source_lines_get_line_number(self):
        to_lines = SourceLinesAdapter("line1\nline2\n")
        self.assertEqual(1, to_lines.get_line_number(0))
        self.assertEqual(1, to_lines.get_line_number(5))
        self.assertEqual(2, to_lines.get_line_number(7))
        self.assertEqual(3, to_lines.get_line_number(12))

    def test_source_lines_get_line_start(self):
        to_lines = SourceLinesAdapter("line1\nline2\n")
        self.assertEqual(0, to_lines.get_line_start(1))
        self.assertEqual(6, to_lines.get_line_start(2))
        self.assertEqual(12, to_lines.get_line_start(3))

    def test_source_lines_get_line_end(self):
        to_lines = SourceLinesAdapter("line1\nline2\n")
        self.assertEqual(5, to_lines.get_line_end(1))
        self.assertEqual(11, to_lines.get_line_end(2))
        self.assertEqual(12, to_lines.get_line_end(3))

    def test_source_lines_last_line_with_no_new_line(self):
        to_lines = SourceLinesAdapter("line1")
        self.assertEqual(1, to_lines.get_line_number(5))


class WordRangeFinderTest(unittest.TestCase):
    def _find_primary(self, code, offset):
        word_finder = worder.Worder(code)
        return word_finder.get_primary_at(offset)

    def _annotated_code(self, annotated_code):
        """
        Split annotated code into raw code and annotation.

        Odd lines in `annotated_code` is the actual Python code.

        Even lines in `annotated_code` are single-char annotation for the
        previous line.

        The annotation may contain one extra character which annotates the
        newline/end of line character.
        """
        code_lines = annotated_code.splitlines()[::2]
        annotations_lines = annotated_code.splitlines()[1::2]
        if len(annotations_lines) < len(code_lines):
            annotations_lines.append("")
        for idx, (line, line_ann) in enumerate(zip(code_lines, annotations_lines)):
            newline_ann_char = 1  # for annotation of the end of line character
            self.assertLessEqual(
                len(line_ann),
                len(line) + newline_ann_char,
                msg="Extra character in annotations",
            )
            line_ann = line_ann.rstrip()
            line_ann += " " * (len(line) - len(line_ann))
            if len(line_ann) != len(line) + newline_ann_char:
                line_ann += " "
            self.assertEqual(len(line_ann), len(line) + newline_ann_char)
            annotations_lines[idx] = line_ann
        code, annotations = "\n".join(code_lines), "\n".join(annotations_lines)
        if code[-1] != "\n":
            annotations = annotations[:-1]
        self.assertEqual(len(code) + code.count("\n"), len(annotations))
        return code, annotations

    def _make_offset_annotation(self, code, func):
        """
        Create annotation by calling `func(offset)` for every offset in `code`.

        For example, when the annotated code looks like so:

            import a.b.c.d
                   ++++++++

        This means that `func(offset)` returns True whenever offset points to
        the 'a.b.c.d' part and returns False everywhere else.
        """

        def _annotation_char(offset):
            ann_char = "+" if func(offset) else " "
            if code[offset] == "\n":
                ann_char = ann_char + "\n"
            return ann_char

        return "".join([_annotation_char(offset) for offset in range(len(code))])

    def assert_equal_annotation(self, code, expected, actual):
        if expected != actual:
            msg = ["Annotation does not match:\n"]
            for line, line_exp, line_actual in zip(
                code.splitlines(), expected.splitlines(), actual.splitlines()
            ):
                msg.append("  " + line + "\n")
                if line_exp != line_actual:
                    msg.append("e " + line_exp + "\n")
                    msg.append("a " + line_actual + "\n")
            self.fail("".join(msg))

    def test_keyword_before_parens(self):
        code = dedent("""\
            if (a_var).an_attr:
                pass
        """)
        self.assertEqual("(a_var).an_attr", self._find_primary(code, code.index(":")))

    def test_inside_parans(self):
        code = "a_func(a_var)"
        self.assertEqual("a_var", self._find_primary(code, 10))

    def test_simple_names(self):
        code = "a_var = 10"
        self.assertEqual("a_var", self._find_primary(code, 3))

    def test_function_calls(self):
        code = "sample_function()"
        self.assertEqual("sample_function", self._find_primary(code, 10))

    def test_attribute_accesses(self):
        code = "a_var.an_attr"
        self.assertEqual("a_var.an_attr", self._find_primary(code, 10))

    def test_word_finder_on_word_beginning(self):
        code = "print(a_var)\n"
        word_finder = worder.Worder(code)
        result = word_finder.get_word_at(code.index("a_var"))
        self.assertEqual("a_var", result)

    def test_word_finder_on_primary_beginning(self):
        code = "print(a_var)\n"
        result = self._find_primary(code, code.index("a_var"))
        self.assertEqual("a_var", result)

    def test_word_finder_on_word_ending(self):
        code = "print(a_var)\n"
        word_finder = worder.Worder(code)
        result = word_finder.get_word_at(code.index("a_var") + 5)
        self.assertEqual("a_var", result)

    def test_word_finder_on_primary_ending(self):
        code = "print(a_var)\n"
        result = self._find_primary(code, code.index("a_var") + 5)
        self.assertEqual("a_var", result)

    def test_word_finder_on_primaries_with_dots_inside_parens(self):
        code = "(a_var.\nattr)"
        result = self._find_primary(code, code.index("attr") + 1)
        self.assertEqual("a_var.\nattr", result)

    def test_word_finder_on_primary_like_keyword(self):
        code = "is_keyword = False\n"
        result = self._find_primary(code, 1)
        self.assertEqual("is_keyword", result)

    def test_keyword_before_parens_no_space(self):
        code = dedent("""\
            if(a_var).an_attr:
                pass
        """)
        self.assertEqual("(a_var).an_attr", self._find_primary(code, code.index(":")))

    def test_strings(self):
        code = '"a string".split()'
        self.assertEqual('"a string".split', self._find_primary(code, 14))

    def test_function_calls2(self):
        code = 'file("afile.txt").read()'
        self.assertEqual('file("afile.txt").read', self._find_primary(code, 18))

    def test_parens(self):
        code = '("afile.txt").split()'
        self.assertEqual('("afile.txt").split', self._find_primary(code, 18))

    def test_function_with_no_param(self):
        code = "AClass().a_func()"
        self.assertEqual("AClass().a_func", self._find_primary(code, 12))

    def test_function_with_multiple_param(self):
        code = 'AClass(a_param, another_param, "a string").a_func()'
        self.assertEqual(
            'AClass(a_param, another_param, "a string").a_func',
            self._find_primary(code, 44),
        )

    def test_param_expressions(self):
        code = "AClass(an_object.an_attr).a_func()"
        self.assertEqual("an_object.an_attr", self._find_primary(code, 20))

    def test_string_parens(self):
        code = 'a_func("(").an_attr'
        self.assertEqual('a_func("(").an_attr', self._find_primary(code, 16))

    def test_extra_spaces(self):
        code = 'a_func  (  "(" ) .   an_attr'
        self.assertEqual('a_func  (  "(" ) .   an_attr', self._find_primary(code, 26))

    def test_relative_import(self):
        code = "from .module import smt"
        self.assertEqual(".module", self._find_primary(code, 5))

    def test_functions_on_ending_parens(self):
        code = "A()"
        self.assertEqual("A()", self._find_primary(code, 2))

    def test_split_statement(self):
        word_finder = worder.Worder("an_object.an_attr")
        self.assertEqual(
            ("an_object", "an_at", 10), word_finder.get_splitted_primary_before(15)
        )

    def test_empty_split_statement(self):
        word_finder = worder.Worder("an_attr")
        self.assertEqual(("", "an_at", 0), word_finder.get_splitted_primary_before(5))

    def test_empty_split_statement2(self):
        word_finder = worder.Worder("an_object.")
        self.assertEqual(
            ("an_object", "", 10), word_finder.get_splitted_primary_before(10)
        )

    def test_empty_split_statement3(self):
        word_finder = worder.Worder("")
        self.assertEqual(("", "", 0), word_finder.get_splitted_primary_before(0))

    def test_empty_split_statement4(self):
        word_finder = worder.Worder("a_var = ")
        self.assertEqual(("", "", 8), word_finder.get_splitted_primary_before(8))

    def test_empty_split_statement5(self):
        word_finder = worder.Worder("a.")
        self.assertEqual(("a", "", 2), word_finder.get_splitted_primary_before(2))

    def test_operators_inside_parens(self):
        code = "(a_var + another_var).reverse()"
        self.assertEqual("(a_var + another_var).reverse", self._find_primary(code, 25))

    def test_dictionaries(self):
        code = 'print({1: "one", 2: "two"}.keys())'
        self.assertEqual('{1: "one", 2: "two"}.keys', self._find_primary(code, 29))

    def test_following_parens(self):
        code = "a_var = a_func()()"
        result = self._find_primary(code, code.index(")(") + 3)
        self.assertEqual("a_func()()", result)

    def test_comments_for_finding_statements(self):
        code = "# var2 . \n  var3"
        self.assertEqual("var3", self._find_primary(code, code.index("3")))

    def test_str_in_comments_for_finding_statements(self):
        code = '# "var2" . \n  var3'
        self.assertEqual("var3", self._find_primary(code, code.index("3")))

    def test_comments_for_finding_statements2(self):
        code = 'var1 + "# var2".\n  var3'
        self.assertEqual("var3", self._find_primary(code, 21))

    def test_comments_for_finding_statements3(self):
        code = '"" + # var2.\n  var3'
        self.assertEqual("var3", self._find_primary(code, 21))

    def test_is_import_statement(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            import a.b.c.d
                   ++++++++
            from a.b import c

            import a.b.c.d as d
                   +++++++++++++
            from a.b import c as e

            from a.b import (

                abc

            )

            result = a.b.c.d.f()

        """))
        word_finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(code, word_finder.is_import_statement),
        )

    def test_is_import_statement_finding(self):
        code = dedent("""\
            import mod
            a_var = 10
        """)
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_import_statement(code.index("mod") + 1))
        self.assertFalse(word_finder.is_import_statement(code.index("a_var") + 1))

    def test_is_import_statement_finding2(self):
        code = dedent("""\
            import a.b.c.d
            result = a.b.c.d.f()
        """)
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_import_statement(code.rindex("d") + 1))

    def test_word_parens_range(self):
        code = dedent("""\
            s = str()
            s.title()
        """)
        word_finder = worder.Worder(code)
        result = word_finder.get_word_parens_range(code.rindex("()") - 1)
        self.assertEqual((len(code) - 3, len(code) - 1), result)

    def test_getting_primary_before_get_index(self):
        code = "\na = (b + c).d[0]()\n"
        result = self._find_primary(code, len(code) - 2)
        self.assertEqual("(b + c).d[0]()", result)

    def test_getting_primary_and_strings_at_the_end_of_line(self):
        code = "f('\\'')\n"
        result = self._find_primary(code, len(code) - 1)  # noqa

    def test_getting_primary_and_not_crossing_newlines(self):
        code = "\na = (b + c)\n(4 + 1).x\n"
        result = self._find_primary(code, len(code) - 1)
        self.assertEqual("(4 + 1).x", result)

    # XXX: concatenated string literals
    def xxx_test_getting_primary_cancatenating_strs(self):
        code = 's = "a"\n"b" "c"\n'
        result = self._find_primary(code, len(code) - 2)
        self.assertEqual('"b" "c"', result)

    def test_is_a_function_being_called_with_parens_on_next_line(self):
        code = "func\n(1, 2)\n"
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_a_function_being_called(1))

    # XXX: handling triple quotes
    def xxx_test_triple_quotes(self):
        code = 's = """string"""\n'
        result = self._find_primary(code, len(code) - 1)
        self.assertEqual('"""string"""', result)

    def test_triple_quotes_spanning_multiple_lines(self):
        code = 's = """\\\nl1\nl2\n """\n'
        result = self._find_primary(code, len(code) - 2)
        self.assertEqual('"""\\\nl1\nl2\n """', result)

    def test_get_word_parens_range_and_string_literals(self):
        code = 'f(1, ")", 2)\n'
        word_finder = worder.Worder(code)
        result = word_finder.get_word_parens_range(0)
        self.assertEqual((1, len(code) - 1), result)

    def test_is_assigned_here_for_equality_test(self):
        code = "a == 1\n"
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_assigned_here(0))

    def test_is_assigned_here_for_not_equal_test(self):
        code = "a != 1\n"
        word_finder = worder.Worder(code)
        self.assertFalse(word_finder.is_assigned_here(0))

    # XXX: is_assigned_here should work for tuple assignments
    def xxx_test_is_assigned_here_for_tuple_assignment(self):
        code = "a, b = (1, 2)\n"
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_assigned_here(0))

    def test_is_from_statement(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            import a.b.c.d

            from a.b import c
                 +++++++++++++
            import a.b.c.d as d

            from a.b import c as e
                 ++++++++++++++++++
            from a.b import (
                 +++++++++++++
                abc
            ++++++++
            )
            ++
            result = a.b.c.d.f()

        """))
        word_finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(code, word_finder.is_from_statement),
        )

    def test_is_from_statement_module(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            import a.b.c.d

            from a.b import c
                +++++
            import a.b.c.d as d

            from a.b import c as e
                +++++
            from a.b import (
                +++++
                abc

            )

            result = a.b.c.d.f()

        """))
        word_finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(code, word_finder.is_from_statement_module),
        )

    def test_is_import_statement_aliased_module(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            import a.b.c.d

            from a.b import c

            import a.b.c.d as d
                   +++++++
            from a.b import c as e

            from a.b import (

                abc

            )

            import mod1, \\

                mod2 as c, mod3, mod4 as d
               +++++            +++++
            result = a.b.c.d.f()

        """))
        word_finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(
                code, word_finder.is_import_statement_aliased_module
            ),
        )

    def test_is_from_aliased(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            import a.b.c.d

            from a.b import c

            import a.b.c.d as d

            from a.b import c as e
                           ++
            from a.b import (

                abc

            )

            from a.b import mod1, \\

                mod2 as c, mod3, mod4 as d
               +++++            +++++
            result = a.b.c.d.f()

        """))
        word_finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(code, word_finder.is_from_aliased),
        )

    def test_is_from_with_from_import_and_multiline_parens(self):
        code = "from mod import \\\n  (f,\n  g, h)\n"
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_from_statement(code.rindex("g")))

    def test_is_from_with_from_import_and_line_breaks_in_the_middle(self):
        code = "from mod import f,\\\n g\n"
        word_finder = worder.Worder(code)
        self.assertTrue(word_finder.is_from_statement(code.rindex("g")))

    def test_is_function_keyword_parameter(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            func(param=1)
                ++++++
            func(

                param=1
               ++++++
            )

            def func(param=1):
                    ++++++
                pass

        """))
        word_finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(
                code, word_finder.is_function_keyword_parameter
            ),
        )

    def test_one_letter_is_function_keyword_parameter(self):
        code = "f(p=1)\n"
        word_finder = worder.Worder(code)
        index = code.rindex("p")
        self.assertTrue(word_finder.is_function_keyword_parameter(index))

    def test_find_parens_start(self):
        code = "f(p)\n"
        finder = worder.Worder(code)
        self.assertEqual(1, finder.find_parens_start_from_inside(2))

    def test_underlined_find_parens_start(self):
        code = 'f(p="")\n'
        finder = worder.Worder(code)
        self.assertEqual(1, finder._find_parens_start(len(code) - 2))

    def test_find_parens_start_with_multiple_entries(self):
        code = "myfunc(p1, p2, p3\n"
        finder = worder.Worder(code)
        self.assertEqual(
            code.index("("), finder.find_parens_start_from_inside(len(code) - 1)
        )

    def test_find_parens_start_with_nested_parens(self):
        code = "myfunc(p1, (p2, p3), p4\n"
        finder = worder.Worder(code)
        self.assertEqual(
            code.index("("), finder.find_parens_start_from_inside(len(code) - 1)
        )

    def test_find_parens_start_with_parens_in_strs(self):
        code = 'myfunc(p1, "(", p4\n'
        finder = worder.Worder(code)
        self.assertEqual(
            code.index("("), finder.find_parens_start_from_inside(len(code) - 1)
        )

    def test_find_parens_start_with_parens_in_strs_in_multiple_lines(self):
        code = 'myfunc  (\np1\n , \n "(" \n, \np4\n'
        finder = worder.Worder(code)
        self.assertEqual(
            code.index("("), finder.find_parens_start_from_inside(len(code) - 1)
        )

    def test_is_on_function_call_keyword(self):
        code, annotations = self._annotated_code(annotated_code=dedent("""\
            myfunc(va
                  +++
        """))

        finder = worder.Worder(code)
        self.assert_equal_annotation(
            code,
            annotations,
            self._make_offset_annotation(code, finder.is_on_function_call_keyword),
        )

    def test_is_on_function_keyword_partial(self):
        code = "myfunc(va"
        finder = worder.Worder(code)
        self.assertTrue(finder.is_on_function_call_keyword(len(code) - 1))

    def test_get_word_range_with_fstring(self):
        code = dedent('''\
            auth = 8
            my_var = f"some value {auth}"
            print(auth)
            other_val = "some other"''')
        finder = worder.Worder(code)
        self.assertEqual(finder.get_word_range(45), (45, 49))


class ScopeNameFinderTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    # FIXME: in normal scopes the interpreter raises `UnboundLocalName`
    # exception, but not in class bodies
    def xxx_test_global_name_in_class_body(self):
        code = dedent("""\
            a_var = 10
            class C(object):
                a_var = a_var
        """)
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        result = name_finder.get_pyname_at(len(code) - 3)
        self.assertEqual(scope["a_var"], result)

    def test_class_variable_attribute_in_class_body(self):
        code = dedent("""\
            a_var = 10
            class C(object):
                a_var = a_var
        """)
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_var_pyname = scope["C"].get_object()["a_var"]
        result = name_finder.get_pyname_at(len(code) - 12)
        self.assertEqual(a_var_pyname, result)

    def test_class_variable_attribute_in_class_body2(self):
        code = dedent("""\
            a_var = 10
            class C(object):
                a_var \\
            = a_var
        """)
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_var_pyname = scope["C"].get_object()["a_var"]
        result = name_finder.get_pyname_at(len(code) - 12)
        self.assertEqual(a_var_pyname, result)

    def test_class_method_attribute_in_class_body(self):
        code = dedent("""\
            class C(object):
                def a_method(self):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_method_pyname = scope["C"].get_object()["a_method"]
        result = name_finder.get_pyname_at(code.index("a_method") + 2)
        self.assertEqual(a_method_pyname, result)

    def test_inner_class_attribute_in_class_body(self):
        code = dedent("""\
            class C(object):
                class CC(object):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        a_class_pyname = scope["C"].get_object()["CC"]
        result = name_finder.get_pyname_at(code.index("CC") + 2)
        self.assertEqual(a_class_pyname, result)

    def test_class_method_in_class_body_but_not_indexed(self):
        code = dedent("""\
            class C(object):
                def func(self, func):
                    pass
        """)
        scope = libutils.get_string_scope(self.project, code)
        a_func_pyname = scope.get_scopes()[0].get_scopes()[0]["func"]
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        result = name_finder.get_pyname_at(code.index(", func") + 3)
        self.assertEqual(a_func_pyname, result)

    def test_function_but_not_indexed(self):
        code = dedent("""\
            def a_func(a_func):
                pass
        """)
        scope = libutils.get_string_scope(self.project, code)
        a_func_pyname = scope["a_func"]
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        result = name_finder.get_pyname_at(code.index("a_func") + 3)
        self.assertEqual(a_func_pyname, result)

    def test_modules_after_from_statements(self):
        root_folder = self.project.root
        mod = testutils.create_module(self.project, "mod", root_folder)
        mod.write(dedent("""\
            def a_func():
                pass
        """))
        code = "from mod import a_func\n"
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        mod_pyobject = self.project.get_pymodule(mod)
        found_pyname = name_finder.get_pyname_at(code.index("mod") + 1)
        self.assertEqual(mod_pyobject, found_pyname.get_object())

    def test_renaming_functions_with_from_import_and_parens(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write(dedent("""\
            def afunc():
                pass
        """))
        code = dedent("""\
            from mod1 import (
                afunc as func)
        """)
        scope = libutils.get_string_scope(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(scope.pyobject)
        mod_pyobject = self.project.get_pymodule(mod1)
        afunc = mod_pyobject["afunc"]
        found_pyname = name_finder.get_pyname_at(code.index("afunc") + 1)
        self.assertEqual(afunc.get_object(), found_pyname.get_object())

    def test_relative_modules_after_from_statements(self):
        pkg1 = testutils.create_package(self.project, "pkg1")
        pkg2 = testutils.create_package(self.project, "pkg2", pkg1)
        mod1 = testutils.create_module(self.project, "mod1", pkg1)
        mod2 = testutils.create_module(self.project, "mod2", pkg2)
        mod1.write(dedent("""\
            def a_func():
                pass
        """))
        code = "from ..mod1 import a_func\n"
        mod2.write(code)
        mod2_scope = self.project.get_pymodule(mod2).get_scope()
        name_finder = rope.base.evaluate.ScopeNameFinder(mod2_scope.pyobject)
        mod1_pyobject = self.project.get_pymodule(mod1)
        found_pyname = name_finder.get_pyname_at(code.index("mod1") + 1)
        self.assertEqual(mod1_pyobject, found_pyname.get_object())

    def test_relative_modules_after_from_statements2(self):
        mod1 = testutils.create_module(self.project, "mod1")
        pkg1 = testutils.create_package(self.project, "pkg1")
        pkg2 = testutils.create_package(self.project, "pkg2", pkg1)
        mod2 = testutils.create_module(self.project, "mod2", pkg2)  # noqa
        mod1.write("import pkg1.pkg2.mod2")

        mod1_scope = self.project.get_pymodule(mod1).get_scope()
        name_finder = rope.base.evaluate.ScopeNameFinder(mod1_scope.pyobject)
        pkg2_pyobject = self.project.get_pymodule(pkg2)
        found_pyname = name_finder.get_pyname_at(mod1.read().index("pkg2") + 1)
        self.assertEqual(pkg2_pyobject, found_pyname.get_object())

    def test_get_pyname_at_on_language_keywords(self):
        code = dedent("""\
            def a_func(a_func):
                pass
        """)
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        with self.assertRaises(exceptions.RopeError):
            name_finder.get_pyname_at(code.index("pass"))

    def test_one_liners(self):
        code = dedent("""\
            var = 1
            def f(): var = 2
            print(var)
        """)
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        pyname = name_finder.get_pyname_at(code.rindex("var"))
        self.assertEqual(pymod["var"], pyname)

    def test_one_liners_with_line_breaks(self):
        code = dedent("""\
            var = 1
            def f(
            ): var = 2
            print(var)
        """)
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        pyname = name_finder.get_pyname_at(code.rindex("var"))
        self.assertEqual(pymod["var"], pyname)

    def test_one_liners_with_line_breaks2(self):
        code = dedent("""\
            var = 1
            def f(
            p): var = 2
            print(var)
        """)
        pymod = libutils.get_string_module(self.project, code)
        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)
        pyname = name_finder.get_pyname_at(code.rindex("var"))
        self.assertEqual(pymod["var"], pyname)

    def test_var_in_list_comprehension_differs_from_var_outside(self):
        code = "var = 1\n[var for var in range(1)]\n"
        pymod = libutils.get_string_module(self.project, code)

        name_finder = rope.base.evaluate.ScopeNameFinder(pymod)

        outside_pyname = name_finder.get_pyname_at(code.index("var"))
        inside_pyname = name_finder.get_pyname_at(code.rindex("var"))

        self.assertNotEqual(outside_pyname, inside_pyname)


class LogicalLineFinderTest(unittest.TestCase):
    def _logical_finder(self, code):
        return LogicalLineFinder(SourceLinesAdapter(code))

    def test_normal_lines(self):
        code = "a_var = 10"
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 1), line_finder.logical_line_in(1))

    def test_normal_lines2(self):
        code = dedent("""\
            another = 10
            a_var = 20
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 1), line_finder.logical_line_in(1))
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_implicit_continuation(self):
        code = "a_var = 3 + \\\n    4 + \\\n    5"
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 3), line_finder.logical_line_in(2))

    def test_explicit_continuation(self):
        code = dedent("""\
            print(2)
            a_var = (3 +
                4,
                5)
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 4), line_finder.logical_line_in(2))

    def test_explicit_continuation_comments(self):
        code = "#\na_var = 3\n"
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_multiple_indented_ifs(self):
        code = dedent("""\
            if True:
                if True:
                    if True:
                        pass
                a = 10
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((5, 5), line_finder.logical_line_in(5))

    def test_list_comprehensions_and_fors(self):
        code = dedent("""\
            a_list = [i
                for i in range(10)]
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 2), line_finder.logical_line_in(2))

    def test_generator_expressions_and_fors(self):
        code = dedent("""\
            a_list = (i
                for i in range(10))
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 2), line_finder.logical_line_in(2))

    def test_fors_and_block_start(self):
        code = dedent("""\
            l = range(10)
            for i in l:
                print(i)
        """)
        self.assertEqual(2, get_block_start(SourceLinesAdapter(code), 2))

    def test_problems_with_inner_indentations(self):
        code = dedent("""\
            if True:
                if True:
                    if True:
                        pass
                a = \\
                    1
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((5, 6), line_finder.logical_line_in(6))

    def test_problems_with_inner_indentations2(self):
        code = dedent("""\
            if True:
                if True:
                    pass
            a = 1
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((4, 4), line_finder.logical_line_in(4))

    def test_logical_lines_for_else(self):
        code = dedent("""\
            if True:
                pass
            else:
                pass
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual((3, 3), line_finder.logical_line_in(3))

    def test_logical_lines_for_lines_with_wrong_continues(self):
        code = "var = 1 + \\"
        line_finder = self._logical_finder(code)
        self.assertEqual((1, 1), line_finder.logical_line_in(1))

    def test_logical_lines_for_multiline_string_with_extra_quotes_front(self):
        code = '""""Docs."""\na = 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_logical_lines_for_multiline_string_with_escaped_quotes(self):
        code = '"""Quotes \\""" "\\"" \' """\na = 1\n'
        line_finder = self._logical_finder(code)
        self.assertEqual((2, 2), line_finder.logical_line_in(2))

    def test_generating_line_starts(self):
        code = dedent("""\
            a = 1
            a = 2

            a = 3
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual([1, 2, 4], list(line_finder.generate_starts()))

    def test_generating_line_starts2(self):
        code = "a = 1\na = 2\n\na = \\ 3\n"
        line_finder = self._logical_finder(code)
        self.assertEqual([2, 4], list(line_finder.generate_starts(2)))

    def test_generating_line_starts3(self):
        code = "a = 1\na = 2\n\na = \\ 3\n"
        line_finder = self._logical_finder(code)
        self.assertEqual([2], list(line_finder.generate_starts(2, 3)))

    def test_generating_line_starts_for_multi_line_statements(self):
        code = "\na = \\\n 1 + \\\n 1\n"
        line_finder = self._logical_finder(code)
        self.assertEqual([2], list(line_finder.generate_starts()))

    def test_generating_line_starts_and_unmatched_deindents(self):
        code = dedent("""\
            if True:
                if True:
                    if True:
                        a = 1
                b = 1
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual([4, 5], list(line_finder.generate_starts(4)))

    def test_false_triple_quoted_string(self):
        code = dedent("""\
            def foo():
                a = 0
                p = 'foo'''

            def bar():
                a = 1
                a += 1
        """)
        line_finder = self._logical_finder(code)
        self.assertEqual([1, 2, 3, 5, 6, 7], list(line_finder.generate_starts()))
        self.assertEqual((3, 3), line_finder.logical_line_in(3))
        self.assertEqual([5, 6, 7], list(line_finder.generate_starts(4)))


class TokenizerLogicalLineFinderTest(LogicalLineFinderTest):
    def _logical_finder(self, code):
        lines = SourceLinesAdapter(code)
        return codeanalyze.CachingLogicalLineFinder(
            lines, codeanalyze.tokenizer_generator
        )


class CustomLogicalLineFinderTest(LogicalLineFinderTest):
    def _logical_finder(self, code):
        lines = SourceLinesAdapter(code)
        return codeanalyze.CachingLogicalLineFinder(lines, codeanalyze.custom_generator)



from __future__ import annotations

import typing

import rope.base.pyobjects
from rope.base import exceptions, utils

if typing.TYPE_CHECKING:
    from typing import Union

    from rope.base import pyobjectsdef


class PyName:
    """References to `PyObject` inside python programs"""

    def get_object(self):
        """Return the `PyObject` object referenced by this `PyName`"""

    def get_definition_location(self):
        """Return a (module, lineno) tuple"""


class DefinedName(PyName):
    def __init__(self, pyobject):
        self.pyobject = pyobject

    def get_object(self):
        return self.pyobject

    def get_definition_location(self):
        lineno = utils.guess_def_lineno(
            self.pyobject.get_module(), self.pyobject.get_ast()
        )
        return (self.pyobject.get_module(), lineno)


class AssignedName(PyName):
    pass


class UnboundName(PyName):
    def __init__(self, pyobject=None):
        self.pyobject = pyobject
        if self.pyobject is None:
            self.pyobject = rope.base.pyobjects.get_unknown()

    def get_object(self):
        return self.pyobject

    def get_definition_location(self):
        return (None, None)


class AssignmentValue:
    """An assigned expression"""

    def __init__(
        self, ast_node, levels=None, evaluation="", assign_type=False, type_hint=None
    ):
        """The `level` is `None` for simple assignments and is
        a list of numbers for tuple assignments for example in::

           a, (b, c) = x

        The levels for for `a` is ``[0]``, for `b` is ``[1, 0]`` and for
        `c` is ``[1, 1]``.

        """
        self.ast_node = ast_node
        if levels is None:
            self.levels = []
        else:
            self.levels = levels
        self.evaluation = evaluation
        self.assign_type = assign_type
        self.type_hint = type_hint

    def get_lineno(self):
        return self.ast_node.lineno


class EvaluatedName(PyName):
    """A name whose object will be evaluated later"""

    def __init__(self, callback, module=None, lineno=None):
        self.module = module
        self.lineno = lineno
        self.callback = callback
        self.pyobject = _Inferred(callback, _get_concluded_data(module))

    def get_object(self):
        return self.pyobject.get()

    def get_definition_location(self):
        return (self.module, self.lineno)

    def invalidate(self):
        """Forget the `PyObject` this `PyName` holds"""
        self.pyobject.set(None)


class ParameterName(PyName):
    pass


class ImportedModule(PyName):
    def __init__(
        self,
        importing_module: Union[
            pyobjectsdef.PyModule,
            pyobjectsdef.PyPackage,
        ],
        module_name=None,
        level=0,
        resource=None,
    ):
        assert (
            module_name is not None or resource is not None
        ), "At least one of module_name or resource must be set"
        self.importing_module = importing_module
        self.module_name = module_name
        self.level = level
        self.resource = resource
        self.pymodule = _get_concluded_data(self.importing_module)

    def _current_folder(self):
        resource = self.importing_module.get_module().get_resource()
        if resource is None:
            return None
        return resource.parent

    def _get_pymodule(self):
        if self.pymodule.get() is None:
            pycore = self.importing_module.pycore
            if self.resource is not None:
                self.pymodule.set(pycore.project.get_pymodule(self.resource))
            else:
                try:
                    if self.level == 0:
                        pymodule = pycore.project.get_module(
                            self.module_name, self._current_folder()
                        )
                    else:
                        pymodule = pycore.project.get_relative_module(
                            self.module_name, self._current_folder(), self.level
                        )
                    self.pymodule.set(pymodule)
                except exceptions.ModuleNotFoundError:
                    pass
        return self.pymodule.get()

    def get_object(self):
        if self._get_pymodule() is None:
            return rope.base.pyobjects.get_unknown()
        return self._get_pymodule()

    def get_definition_location(self):
        pymodule = self._get_pymodule()
        if not isinstance(pymodule, rope.base.pyobjects.PyDefinedObject):
            return (None, None)
        return (pymodule.get_module(), 1)


class ImportedName(PyName):
    def __init__(self, imported_module, imported_name):
        self.imported_module = imported_module
        self.imported_name = imported_name

    def _get_imported_pyname(self):
        try:
            result = self.imported_module.get_object()[self.imported_name]
            if result != self:
                return result
        except exceptions.AttributeNotFoundError:
            pass
        return UnboundName()

    @utils.prevent_recursion(rope.base.pyobjects.get_unknown)
    def get_object(self):
        return self._get_imported_pyname().get_object()

    @utils.prevent_recursion(lambda: (None, None))
    def get_definition_location(self):
        return self._get_imported_pyname().get_definition_location()


def _get_concluded_data(module):
    if module is None:
        return rope.base.pyobjects._ConcludedData()
    return module._get_concluded_data()


def _circular_inference():
    raise rope.base.pyobjects.IsBeingInferredError("Circular Object Inference")


class _Inferred:
    def __init__(self, get_inferred, concluded=None):
        self.get_inferred = get_inferred
        self.concluded = concluded
        if self.concluded is None:
            self.temp = None

    @utils.prevent_recursion(_circular_inference)
    def get(self, *args, **kwds):
        if self.concluded is None or self.concluded.get() is None:
            self.set(self.get_inferred(*args, **kwds))
        if self._get() is None:
            self.set(rope.base.pyobjects.get_unknown())
        return self._get()

    def set(self, pyobject):
        if self.concluded is not None:
            self.concluded.set(pyobject)
        self.temp = pyobject

    def _get(self):
        if self.concluded is not None:
            return self.concluded.get()
        return self.temp



import unittest

from rope.contrib.fixmodnames import FixModuleNames
from rope.contrib.generate import create_module, create_package
from ropetest import testutils


# HACK: for making this test work on case-insensitive file-systems, it
# uses a name.replace('x', '_') fixer.
class FixModuleNamesTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_simple_module_renaming(self):
        mod = create_module(self.project, "xod")
        self.project.do(FixModuleNames(self.project).get_changes(_fixer))
        self.assertFalse(mod.exists())
        self.assertTrue(self.project.get_resource("_od.py").exists())

    def test_packages_module_renaming(self):
        pkg = create_package(self.project, "xkg")
        self.project.do(FixModuleNames(self.project).get_changes(_fixer))
        self.assertFalse(pkg.exists())
        self.assertTrue(self.project.get_resource("_kg/__init__.py").exists())

    def test_fixing_contents(self):
        mod1 = create_module(self.project, "xod1")
        mod2 = create_module(self.project, "xod2")
        mod1.write("import xod2\n")
        mod2.write("import xod1\n")
        self.project.do(FixModuleNames(self.project).get_changes(_fixer))
        newmod1 = self.project.get_resource("_od1.py")
        newmod2 = self.project.get_resource("_od2.py")
        self.assertEqual("import _od2\n", newmod1.read())
        self.assertEqual("import _od1\n", newmod2.read())

    def test_handling_nested_modules(self):
        pkg = create_package(self.project, "xkg")
        mod = create_module(self.project, "xkg.xod")  # noqa
        self.project.do(FixModuleNames(self.project).get_changes(_fixer))
        self.assertFalse(pkg.exists())
        self.assertTrue(self.project.get_resource("_kg/__init__.py").exists())
        self.assertTrue(self.project.get_resource("_kg/_od.py").exists())


def _fixer(name):
    return name.replace("x", "_")



import copy

import rope.base.exceptions
from rope.base import codeanalyze, evaluate, pyobjects, taskhandle, utils, worder
from rope.base.change import ChangeContents, ChangeSet
from rope.refactor import functionutils, occurrences


class ChangeSignature:
    def __init__(self, project, resource, offset):
        self.project = project
        self.resource = resource
        self.offset = offset
        self._set_name_and_pyname()
        if (
            self.pyname is None
            or self.pyname.get_object() is None
            or not isinstance(self.pyname.get_object(), pyobjects.PyFunction)
        ):
            raise rope.base.exceptions.RefactoringError(
                "Change method signature should be performed on functions"
            )

    def _set_name_and_pyname(self):
        self.name = worder.get_name_at(self.resource, self.offset)
        this_pymodule = self.project.get_pymodule(self.resource)
        self.primary, self.pyname = evaluate.eval_location2(this_pymodule, self.offset)
        if self.pyname is None:
            return
        pyobject = self.pyname.get_object()
        if isinstance(pyobject, pyobjects.PyClass) and "__init__" in pyobject:
            self.pyname = pyobject["__init__"]
            self.name = "__init__"
        pyobject = self.pyname.get_object()
        self.others = None
        if (
            self.name == "__init__"
            and isinstance(pyobject, pyobjects.PyFunction)
            and isinstance(pyobject.parent, pyobjects.PyClass)
        ):
            pyclass = pyobject.parent
            self.others = (pyclass.get_name(), pyclass.parent[pyclass.get_name()])

    def _change_calls(
        self,
        call_changer,
        in_hierarchy=None,
        resources=None,
        handle=taskhandle.DEFAULT_TASK_HANDLE,
    ):
        if resources is None:
            resources = self.project.get_python_files()
        changes = ChangeSet("Changing signature of <%s>" % self.name)
        job_set = handle.create_jobset("Collecting Changes", len(resources))
        finder = occurrences.create_finder(
            self.project,
            self.name,
            self.pyname,
            instance=self.primary,
            in_hierarchy=in_hierarchy and self.is_method(),
        )
        if self.others:
            name, pyname = self.others
            constructor_finder = occurrences.create_finder(
                self.project, name, pyname, only_calls=True
            )
            finder = _MultipleFinders([finder, constructor_finder])
        for file in resources:
            job_set.started_job(file.path)
            change_calls = _ChangeCallsInModule(
                self.project, finder, file, call_changer
            )
            changed_file = change_calls.get_changed_module()
            if changed_file is not None:
                changes.add_change(ChangeContents(file, changed_file))
            job_set.finished_job()
        return changes

    def get_args(self):
        """Get function arguments.

        Return a list of ``(name, default)`` tuples for all but star
        and double star arguments.  For arguments that don't have a
        default, `None` will be used.
        """
        return self._definfo().args_with_defaults

    def is_method(self):
        pyfunction = self.pyname.get_object()
        return isinstance(pyfunction.parent, pyobjects.PyClass)

    @utils.deprecated("Use `ChangeSignature.get_args()` instead")
    def get_definition_info(self):
        return self._definfo()

    def _definfo(self):
        return functionutils.DefinitionInfo.read(self.pyname.get_object())

    @utils.deprecated()
    def normalize(self):
        changer = _FunctionChangers(
            self.pyname.get_object(), self.get_definition_info(), [ArgumentNormalizer()]
        )
        return self._change_calls(changer)

    @utils.deprecated()
    def remove(self, index):
        changer = _FunctionChangers(
            self.pyname.get_object(),
            self.get_definition_info(),
            [ArgumentRemover(index)],
        )
        return self._change_calls(changer)

    @utils.deprecated()
    def add(self, index, name, default=None, value=None):
        changer = _FunctionChangers(
            self.pyname.get_object(),
            self.get_definition_info(),
            [ArgumentAdder(index, name, default, value)],
        )
        return self._change_calls(changer)

    @utils.deprecated()
    def inline_default(self, index):
        changer = _FunctionChangers(
            self.pyname.get_object(),
            self.get_definition_info(),
            [ArgumentDefaultInliner(index)],
        )
        return self._change_calls(changer)

    @utils.deprecated()
    def reorder(self, new_ordering):
        changer = _FunctionChangers(
            self.pyname.get_object(),
            self.get_definition_info(),
            [ArgumentReorderer(new_ordering)],
        )
        return self._change_calls(changer)

    def get_changes(
        self,
        changers,
        in_hierarchy=False,
        resources=None,
        task_handle=taskhandle.DEFAULT_TASK_HANDLE,
    ):
        """Get changes caused by this refactoring

        `changers` is a list of `_ArgumentChanger`.  If `in_hierarchy`
        is `True` the changers are applied to all matching methods in
        the class hierarchy.
        `resources` can be a list of `rope.base.resource.File` that
        should be searched for occurrences; if `None` all python files
        in the project are searched.

        """
        function_changer = _FunctionChangers(
            self.pyname.get_object(), self._definfo(), changers
        )
        return self._change_calls(
            function_changer, in_hierarchy, resources, task_handle
        )


class _FunctionChangers:
    def __init__(self, pyfunction, definition_info, changers=None):
        self.pyfunction = pyfunction
        self.definition_info = definition_info
        self.changers = changers
        self.changed_definition_infos = self._get_changed_definition_infos()

    def _get_changed_definition_infos(self):
        definition_info = self.definition_info
        result = [definition_info]
        for changer in self.changers:
            definition_info = copy.deepcopy(definition_info)
            changer.change_definition_info(definition_info)
            result.append(definition_info)
        return result

    def change_definition(self, call):
        return self.changed_definition_infos[-1].to_string()

    def change_call(self, primary, pyname, call):
        call_info = functionutils.CallInfo.read(
            primary, pyname, self.definition_info, call
        )
        mapping = functionutils.ArgumentMapping(self.definition_info, call_info)

        for definition_info, changer in zip(
            self.changed_definition_infos, self.changers
        ):
            changer.change_argument_mapping(definition_info, mapping)

        return mapping.to_call_info(self.changed_definition_infos[-1]).to_string()


class _ArgumentChanger:
    def change_definition_info(self, definition_info):
        pass

    def change_argument_mapping(self, definition_info, argument_mapping):
        pass


class ArgumentNormalizer(_ArgumentChanger):
    pass


class ArgumentRemover(_ArgumentChanger):
    def __init__(self, index):
        self.index = index

    def change_definition_info(self, call_info):
        if self.index < len(call_info.args_with_defaults):
            del call_info.args_with_defaults[self.index]
        elif (
            self.index == len(call_info.args_with_defaults)
            and call_info.args_arg is not None
        ):
            call_info.args_arg = None
        elif (
            self.index == len(call_info.args_with_defaults)
            and call_info.args_arg is None
            and call_info.keywords_arg is not None
        ) or (
            self.index == len(call_info.args_with_defaults) + 1
            and call_info.args_arg is not None
            and call_info.keywords_arg is not None
        ):
            call_info.keywords_arg = None

    def change_argument_mapping(self, definition_info, mapping):
        if self.index < len(definition_info.args_with_defaults):
            name = definition_info.args_with_defaults[0]
            if name in mapping.param_dict:
                del mapping.param_dict[name]


class ArgumentAdder(_ArgumentChanger):
    def __init__(self, index, name, default=None, value=None):
        self.index = index
        self.name = name
        self.default = default
        self.value = value

    def change_definition_info(self, definition_info):
        for pair in definition_info.args_with_defaults:
            if pair[0] == self.name:
                raise rope.base.exceptions.RefactoringError(
                    "Adding duplicate parameter: <%s>." % self.name
                )
        definition_info.args_with_defaults.insert(self.index, (self.name, self.default))

    def change_argument_mapping(self, definition_info, mapping):
        if self.value is not None:
            mapping.param_dict[self.name] = self.value


class ArgumentDefaultInliner(_ArgumentChanger):
    def __init__(self, index):
        self.index = index
        self.remove = False

    def change_definition_info(self, definition_info):
        if self.remove:
            definition_info.args_with_defaults[self.index] = (
                definition_info.args_with_defaults[self.index][0],
                None,
            )

    def change_argument_mapping(self, definition_info, mapping):
        default = definition_info.args_with_defaults[self.index][1]
        name = definition_info.args_with_defaults[self.index][0]
        if default is not None and name not in mapping.param_dict:
            mapping.param_dict[name] = default


class ArgumentReorderer(_ArgumentChanger):
    def __init__(self, new_order, autodef=None):
        """Construct an `ArgumentReorderer`

        Note that the `new_order` is a list containing the new
        position of parameters; not the position each parameter
        is going to be moved to. (changed in ``0.5m4``)

        For example changing ``f(a, b, c)`` to ``f(c, a, b)``
        requires passing ``[2, 0, 1]`` and *not* ``[1, 2, 0]``.

        The `autodef` (automatic default) argument, forces rope to use
        it as a default if a default is needed after the change.  That
        happens when an argument without default is moved after
        another that has a default value.  Note that `autodef` should
        be a string or `None`; the latter disables adding automatic
        default.

        """
        self.new_order = new_order
        self.autodef = autodef

    def change_definition_info(self, definition_info):
        new_args = list(definition_info.args_with_defaults)
        for new_index, index in enumerate(self.new_order):
            new_args[new_index] = definition_info.args_with_defaults[index]
        seen_default = False
        for index, (arg, default) in enumerate(list(new_args)):
            if default is not None:
                seen_default = True
            if seen_default and default is None and self.autodef is not None:
                new_args[index] = (arg, self.autodef)
        definition_info.args_with_defaults = new_args


class _ChangeCallsInModule:
    def __init__(self, project, occurrence_finder, resource, call_changer):
        self.project = project
        self.occurrence_finder = occurrence_finder
        self.resource = resource
        self.call_changer = call_changer

    def get_changed_module(self):
        word_finder = worder.Worder(self.source)
        change_collector = codeanalyze.ChangeCollector(self.source)
        for occurrence in self.occurrence_finder.find_occurrences(self.resource):
            if not occurrence.is_called() and not occurrence.is_defined():
                continue
            start, end = occurrence.get_primary_range()
            begin_parens, end_parens = word_finder.get_word_parens_range(end - 1)
            if occurrence.is_called():
                primary, pyname = occurrence.get_primary_and_pyname()
                changed_call = self.call_changer.change_call(
                    primary, pyname, self.source[start:end_parens]
                )
            else:
                changed_call = self.call_changer.change_definition(
                    self.source[start:end_parens]
                )
            if changed_call is not None:
                change_collector.add_change(start, end_parens, changed_call)
        return change_collector.get_changed()

    @property
    @utils.saveit
    def pymodule(self):
        return self.project.get_pymodule(self.resource)

    @property
    @utils.saveit
    def source(self):
        if self.resource is not None:
            return self.resource.read()
        else:
            return self.pymodule.source_code

    @property
    @utils.saveit
    def lines(self):
        return self.pymodule.lines


class _MultipleFinders:
    def __init__(self, finders):
        self.finders = finders

    def find_occurrences(self, resource=None, pymodule=None):
        all_occurrences = []
        for finder in self.finders:
            all_occurrences.extend(finder.find_occurrences(resource, pymodule))
        all_occurrences.sort(key=lambda x: x.get_primary_range())
        return all_occurrences



import json

import pytest

from rope.base.serializer import _js2py, json_to_python, python_to_json


@pytest.mark.parametrize("version", [1, 2])
@pytest.mark.parametrize(
    "original_data",
    [
        None,
        4,
        "3",
        (),
        [],
        ("hello",),
        (1, [2], "hello"),
        [1, (2,), "hello"],
        {"hello": "world"},
        {"hello": ("hello", 1)},
        {("hello", 1): "world"},
        {"4": "hello"},
        {4: "hello"},
    ],
)
def test_roundtrip(original_data, version):
    encoded = python_to_json(original_data, version)
    serialized = json.dumps(encoded)
    decoded = json.loads(serialized)
    rehydrated_data = json_to_python(decoded)

    assert encoded == decoded
    assert rehydrated_data == original_data


def expected_encoded_simple_data(version):
    data = [
        (None, None),
        (4, 4),
        ("3", "3"),
        ({"hello": "world"}, {"hello": "world"}),
    ]
    return [[*d, version] for d in data]

@pytest.mark.parametrize(
    "original_data,expected_encoded,version",
    [
        *expected_encoded_simple_data(version=1),
        *expected_encoded_simple_data(version=2),

        ((), {"$": "t", "items": []}, 1),
        ([], [], 1),
        (("hello",), {"$": "t", "items": ["hello"]}, 1),
        ((1, [2], "hello"), {"$": "t", "items": [1, [2], "hello"]}, 1),
        ([1, (2,), "hello"], [1, {"$": "t", "items": [2]}, "hello"], 1),
        ({"hello": ("hello", 1)}, {"hello": {"$": "t", "items": ["hello", 1]}}, 1),

        ((), [], 2),
        ([], {"$": "l", "items": []}, 2),
        (("hello",), ["hello"], 2),
        ((1, [2], "hello"), [1, {"$": "l", "items": [2]}, "hello"], 2),
        ([1, (2,), "hello"], {"$": "l", "items": [1, [2], "hello"]}, 2),
        ({"hello": ("hello", 1)}, {"hello": ["hello", 1]}, 2),
    ],
)
def test_expected_encoded_simple(original_data, expected_encoded, version):
    encoded = python_to_json(original_data, version)
    serialized = json.dumps(encoded)
    decoded = json.loads(serialized)
    rehydrated_data = json_to_python(decoded)

    assert encoded == decoded
    assert encoded["v"] == version
    assert encoded["data"] == expected_encoded, (original_data, version)
    assert "references" not in encoded
    assert rehydrated_data == original_data


@pytest.mark.parametrize(
    "original_data,expected_encoded,expected_references,version",
    [
        (
            {("hello", 1): "world"},
            {"0": "world"},
            [{"$": "t", "items": ["hello", 1]}],
            1,
        ),
        (
            {"4": "hello"},
            {"0": "hello"},
            ["4"],
            1,
        ),
        (
            {4: "hello"},
            {"0": "hello"},
            [4],
            1,
        ),
        (
            {None: "hello"},
            {"0": "hello"},
            [None],
            1,
        ),
    ],
)
def test_expected_encoded_with_references(original_data, expected_encoded, expected_references, version):
    encoded = python_to_json(original_data, version)
    serialized = json.dumps(encoded)
    decoded = json.loads(serialized)
    rehydrated_data = json_to_python(decoded)

    assert encoded == decoded
    assert encoded["v"] == 1
    assert encoded["data"] == expected_encoded
    assert encoded["references"] == expected_references
    assert rehydrated_data == original_data



@pytest.mark.parametrize("version", [1, 2])
@pytest.mark.parametrize(
    "original_data,exctype",
    [
        (object(), TypeError),
        (4.8, TypeError),
        ({"$": "hello"}, ValueError),
    ],
)
def test_rejects_unrecognized_object(original_data, exctype, version):
    with pytest.raises(exctype):
        python_to_json(original_data, version)


def test_unexpected_version_python_to_json():
    with pytest.raises(ValueError, match="Unexpected version"):
        python_to_json({"hello": ["world"]}, version=-123456)


def test_unexpected_version_json_to_python():
    modified = python_to_json({"hello": ["world"]})
    modified["v"] = -123456
    assert isinstance(modified["data"]["hello"], list)

    with pytest.raises(ValueError, match="Unexpected version"):
        json_to_python(modified)

    with pytest.raises(ValueError, match="Unexpected version"):
        _js2py(modified["data"], {}, modified["v"])


def test_unexpected_dollar_object_type():
    modified = python_to_json({"hello": ["world"]})
    modified["data"]["$"] = "unexpected"

    with pytest.raises(TypeError, match="Unrecognized object of type"):
        json_to_python(modified)


def test_unexpected_object_type():
    modified = python_to_json({"hello": ["world"]})
    modified["data"]["hello"] = ()

    with pytest.raises(TypeError, match='Object of type "tuple" is not allowed'):
        json_to_python(modified)



import unittest

import rope.base.change
import rope.base.history
import rope.contrib.changestack
from ropetest import testutils


class ChangeStackTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_change_stack(self):
        myfile = self.project.root.create_file("myfile.txt")
        myfile.write("1")
        stack = rope.contrib.changestack.ChangeStack(self.project)
        stack.push(rope.base.change.ChangeContents(myfile, "2"))
        self.assertEqual("2", myfile.read())
        stack.push(rope.base.change.ChangeContents(myfile, "3"))
        self.assertEqual("3", myfile.read())
        stack.pop_all()
        self.assertEqual("1", myfile.read())
        changes = stack.merged()
        self.project.do(changes)
        self.assertEqual("3", myfile.read())



from rope.base import ast


def get_name_levels(node):
    """Return a list of ``(name, level)`` tuples for assigned names

    The `level` is `None` for simple assignments and is a list of
    numbers for tuple assignments for example in::

      a, (b, c) = x

    The levels for for `a` is ``[0]``, for `b` is ``[1, 0]`` and for
    `c` is ``[1, 1]``.

    """
    visitor = _NodeNameCollector()
    visitor.visit(node)
    return visitor.names


class _NodeNameCollector(ast.RopeNodeVisitor):
    def __init__(self, levels=None):
        self.names = []
        self.levels = levels
        self.index = 0

    def _add_node(self, node):
        new_levels = []
        if self.levels is not None:
            new_levels = list(self.levels)
            new_levels.append(self.index)
        self.index += 1
        self._added(node, new_levels)

    def _added(self, node, levels):
        if hasattr(node, "id"):
            self.names.append((node.id, levels))

    def _Name(self, node):
        self._add_node(node)

    def _ExceptHandler(self, node):
        self.names.append((node.name, []))

    def _Tuple(self, node):
        new_levels = []
        if self.levels is not None:
            new_levels = list(self.levels)
            new_levels.append(self.index)
        self.index += 1
        visitor = _NodeNameCollector(new_levels)
        for child in ast.iter_child_nodes(node):
            visitor.visit(child)
        self.names.extend(visitor.names)

    def _Subscript(self, node):
        self._add_node(node)

    def _Attribute(self, node):
        self._add_node(node)

    def _Slice(self, node):
        self._add_node(node)
